import autoroot
import os
from loguru import logger as log
import torch
from torch.utils.data import DataLoader
from vrl_config import argument_parser
from models.base_block import *
from dataset.AttrDataset import MultiModalAttrDataset, get_transform
from tools.utils import set_seed
from dataset.preprocess.vrl_pad import attr_words

from clip import clip
from clip.model import *


set_seed(605)
device = "cuda"


def main(args):

    _, valid_tsfm = get_transform(args)
    test_set = MultiModalAttrDataset(
        args=args,
        split=args.valid_split,
        transform=valid_tsfm,
        dataset_dir=args.data_dir,
    )

    valid_loader = DataLoader(
        dataset=test_set,
        batch_size=1,
        shuffle=False,
        num_workers=8,
        pin_memory=True,
    )

    log.info(
        f"{args.valid_split} set: {len(valid_loader.dataset)}, "
        f"attr_num : {test_set.attr_num}"
    )
    
    checkpoint = torch.load(args.dir)
    clip_model = build_model(checkpoint["clip_model"])
    model = TransformerClassifier(clip_model, test_set.attr_num, test_set.attributes)
    model.load_state_dict(checkpoint['model_state_dict'], strict=True)
    model.eval()
    if torch.cuda.is_available():
        model = model.cuda()
        clip_model = clip_model.cuda()

    directory = os.path.dirname(args.dir)
    directory = os.path.join(os.path.dirname(__file__), directory)
    log.info(f"saving outputs to dir: {directory}")
    pred_dir = os.path.join(directory, "preds")
    os.makedirs(pred_dir, exist_ok=True)
    submit_file_path = os.path.join(directory, "submission.txt")
    with open(submit_file_path, "w") as submit_file:
        with torch.no_grad():
            for step, (imgs, _, imgname) in enumerate(valid_loader):
                # im_name = imgname[0].split('.')[0]
                log.info(f"index: {step+1}, imgname: {imgname}")
                imgs = imgs.cuda()
                valid_logits, _ = model(imgs, clip_model=clip_model)

                valid_probs = torch.sigmoid(valid_logits)
                valid_probs = valid_probs.cpu().numpy()
                preds = (valid_probs > 0.45).astype(int)
                submit_file.write(f"{step+1} {' '.join(map(str, preds[0]))}\n")
                individual_preds_path = os.path.join(pred_dir, f"{step+1}.txt")
                attributes = []
                for i in range(len(preds[0])):
                    if preds[0][i] == 1:
                        attributes.append(attr_words[i])
                with open(individual_preds_path, "w") as f:
                    f.write("\n".join(attributes))

    log.info('test inference complete.')


if __name__ == '__main__':
    # python test_example.py VRL --valid_split test --checkpoint --dir logs/VRL/2024-06-07_01_46_29/epoch115.pth --use_div --use_vismask --vis_prompt 50 --use_GL --use_textprompt --use_mm_former
    parser = argument_parser()
    args = parser.parse_args()
    assert args.checkpoint, "checkpoint is required"
    # assert args.valid_split == "test", "valid_split should be test"
    assert args.dir is not None, "dir is required"
    main(args)