import autoroot
import os
from tqdm import tqdm
from loguru import logger as log
import pprint
from datetime import datetime
from collections import OrderedDict
import numpy as np
import torch
from torch.utils.data import DataLoader
import time
import wandb
from batch_engine import valid_trainer, batch_trainer
from vrl_config import argument_parser
from dataset.AttrDataset import MultiModalAttrDataset, get_transform
from loss.CE_loss import *
from models.base_block import *
from tools.function import get_pedestrian_metrics, count_parameters
from tools.utils import time_str, ReDirectSTD, set_seed
from solver.scheduler_factory import create_scheduler, make_scheduler
import torch.optim as optim
from clip import clip
from clip.model import *


set_seed(605)
device = "cuda"


def main(args):
    start_time=time_str()
    log.info(f'start_time is {start_time}')
    log_dir = os.path.join('logs', args.dataset)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir, exist_ok=True)
    log_dir = os.path.join(log_dir, start_time)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir, exist_ok=True)
    stdout_file = os.path.join(log_dir, f'stdout_{time_str()}.txt')

    if args.redirector:
        print('redirector stdout')
        ReDirectSTD(stdout_file, 'stdout', False)

    log.info(pprint.pformat(OrderedDict(args.__dict__)))
    log.info(
        f"{args.dataset}: train set: {args.train_split}, val set: {args.valid_split}"
    )

    wandb.init(
        name=f"VRL_{datetime.now().strftime('%d-%m-%Y_%H-%M-%S')}",
        config=args,
        allow_val_change=True,
        settings=wandb.Settings(start_method="thread"),
        reinit=True,
        # mode='disabled'
    )

    train_tsfm, valid_tsfm = get_transform(args) 
    train_set = MultiModalAttrDataset(
        args=args,
        split=args.train_split,
        transform=train_tsfm,
        dataset_dir=args.data_dir,
    )
    train_loader = DataLoader(
        dataset=train_set,
        batch_size=args.batchsize, 
        shuffle=True,
        num_workers=8,  
        pin_memory=True,  
    )
    valid_set = MultiModalAttrDataset(
        args=args,
        split=args.valid_split,
        transform=valid_tsfm,
        dataset_dir=args.data_dir,
    )
    valid_loader = DataLoader(
        dataset=valid_set,
        batch_size=args.batchsize,
        shuffle=False,
        num_workers=8,
        pin_memory=True,
    )

    log.info(f'{args.train_split} set: {len(train_loader.dataset)}, args.data_dir{args.valid_split} set: {len(valid_loader.dataset)}, attr_num : {train_set.attr_num}')
    
    labels = train_set.label
    sample_weight = labels.mean(0)
    
    if args.dir: # when finetuning
        log.info(f'loading checkpoint from {args.dir}')
        checkpoint = torch.load(args.dir) # path including the .pth checkpoint
        clip_model = build_model(checkpoint["clip_model"])
        clip_model.reset_attr_num(train_set.attr_num)
        clip_model = clip_model.cuda()
        model = TransformerClassifier(clip_model, args.checkpoint_attr, train_set.attributes)
        model.load_state_dict(checkpoint['model_state_dict'], strict=True)
        model.reset_layers(train_set.attr_num)
    else:
        clip_model, _ = clip.load("ViT-L/14", device=device, download_root=os.environ["cache_dir"])
        model = TransformerClassifier(clip_model, train_set.attr_num, train_set.attributes)

    if torch.cuda.is_available():
        model = model.cuda()
    
    mm_params=[]

    #Freeze parameters other than category headers and prompts
    for name, param in model.named_parameters():
        if any(keyword in name for keyword in args.mmformer_update_parameters):
            mm_params+= [{
            "params": [param],
            "lr": args.lr,
            "weight_decay": args.weight_decay
            }]
        else:
            param.requires_grad = False
    clip_params=[]
    for name, param in clip_model.named_parameters():
        if any(keyword in name for keyword in args.clip_update_parameters):
            print(name, param.requires_grad)
            clip_params+= [{
            "params": [param],
            "lr": args.clip_lr,
            "weight_decay": args.clip_weight_decay
            }]
        else:
            param.requires_grad = False

    criterion = CEL_Sigmoid(sample_weight, attr_idx=train_set.attr_num)
    epoch_num = args.epoch                                                                                                                                                                                            
    count_parameters(model,clip_model,args.mmformer_update_parameters,args.clip_update_parameters)
                
    prompt_optimizer = optim.SGD(clip_params, args.clip_lr, momentum=0.9, weight_decay=args.clip_weight_decay)
    prompt_scheduler = make_scheduler(prompt_optimizer,num_epochs=epoch_num,lr=args.clip_lr,warmup_t=10)
    
    optimizer = optim.SGD(mm_params, args.lr, momentum=0.9, weight_decay=args.weight_decay)
    scheduler = create_scheduler(optimizer, num_epochs=epoch_num, lr=args.lr, warmup_t=5)

    trainer(epoch=epoch_num,
            model=model,
            clip_model=clip_model,
            train_loader=train_loader,
            valid_loader=valid_loader,
            criterion=criterion,
            optimizer=optimizer,
            scheduler=scheduler,
            prompt_scheduler=prompt_scheduler,
            prompt_optimizer=prompt_optimizer,
            args=args,
            path=log_dir)


def trainer(epoch, model, clip_model, train_loader, valid_loader, criterion, optimizer, scheduler,prompt_scheduler,prompt_optimizer,args,path):

    max_ma, max_acc, max_f1,= 0, 0, 0
    start=time.time()

    for i in tqdm(range(1, epoch+1)):
        scheduler.step(i)
        prompt_scheduler.step(i)
        train_loss, train_gt, train_probs = batch_trainer(
            epoch=i,
            model=model,
            clip_model=clip_model,
            train_loader=train_loader,
            criterion=criterion,
            optimizer=optimizer,
            prompt_optimizer=prompt_optimizer,
            args=args
        )
        train_result = get_pedestrian_metrics(train_gt, train_probs)
        valid_loss, valid_gt, valid_probs = valid_trainer(
            model=model,
            clip_model=clip_model,
            valid_loader=valid_loader,
            criterion=criterion,
            args=args
        )
        valid_result = get_pedestrian_metrics(valid_gt, valid_probs)
        log.info(
            f"Evaluation on val set, valid_loss:{valid_loss:.4f}\n"
            f"ma: {valid_result.ma:.4f},  pos_recall: {np.mean(valid_result.label_pos_recall):.4f}, neg_recall: {np.mean(valid_result.label_neg_recall):.4f}\n"
            f"Acc: {valid_result.instance_acc:.4f}, Prec: {valid_result.instance_prec:.4f}, Rec: {valid_result.instance_recall:.4f}, F1: {valid_result.instance_f1:.4f}"
        )
        if valid_result.ma>=max_ma :
            max_ma=valid_result.ma
            torch.save(
                {
                    "epoch": i,
                    "optimizer": optimizer.state_dict(),
                    "model_state_dict": model.state_dict(),
                    "clip_model": clip_model.state_dict(),
                    "result": valid_result,
                },
                os.path.join(path, "best_ma.pth"),
            )
            log.info(f"best ma model saved at epoch {i}, with ma: {valid_result.ma:.4f}")
        if valid_result.instance_acc>=max_acc :
            max_acc=valid_result.instance_acc
        if valid_result.instance_f1>=max_f1 :
            max_f1=valid_result.instance_f1

        wandb.log(
            {
                "train/loss": train_loss,
                "train/mA": train_result.ma,
                "train/Accuray": train_result.instance_acc,
                "train/Precision": train_result.instance_prec,
                "train/Recall": train_result.instance_recall,
                "train/F1": train_result.instance_f1,
                "valid/loss": valid_loss,
                "valid/mA": valid_result.ma,
                "valid/Accuray": valid_result.instance_acc,
                "valid/Precision": valid_result.instance_prec,
                "valid/Recall": valid_result.instance_recall,
                "valid/F1": valid_result.instance_f1,
            }
        )

        if args.save_freq > 0 and i % args.save_freq == 0:
            torch.save({
                        'epoch': i,
                        'optimizer':optimizer.state_dict(),
                        'model_state_dict': model.state_dict(),
                        'clip_model' : clip_model.state_dict(),
                        'result': valid_result
                        }, os.path.join(path, f"epoch{i}.pth"))
        if i%5 == 0:
            end=time.time()
            total=end-start
            log.info(f'The time taken for the last {i} epoch is:{total} \n'
                     f'max_ma:{max_ma:.4f}, max_acc:{max_acc:.4f}, max_f1:{max_f1:.4f}')


if __name__ == '__main__':
    parser = argument_parser()
    args = parser.parse_args()
    main(args)
